import time
import typer, glob
from pathlib import Path
from typing_extensions import Annotated
from rich.progress import Progress, SpinnerColumn, TextColumn
from basebio import check_path_exists, minimap2
from ..utils.ployA_tools import convert_to_fast5_with_summary_file, index_fastq, detect_ployA, ployADetector

app = typer.Typer()

@app.command()
def ployA(
    input: Annotated[str, typer.Option("--input", "-i", help="Input fastq file.")],
    transcriptome: Annotated[str, typer.Option("--transcriptome", "-f", help="Reference transcriptome fasta file path.")],
    output: Annotated[Path, typer.Option("--output", "-o", help="Output file path.")],
    prefix: Annotated[str, typer.Option("--prefix", "-p", help="Prefix for output files.")],
    min_a_length: Annotated[int, typer.Option("--min-a-length", "-a", help="Minimum length of ployA tail.")]=6,
    max_non_a: Annotated[int, typer.Option("--max-non-a", "-n", help="Maximum number of non-A characters in ployA tail.")]=3,
    pod5s: Annotated[str, typer.Option("--pod5s", help="Regular matching pattern for pod5 files, such as 'path/to/*pod5'.")]=None, # type: ignore
    threads: Annotated[int, typer.Option("--threads", "-t", help="Number of threads to use.")]=8,
    ):
    """
    Detect ployA with pod5 and fastq files.
    """
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Detect ployA start...", total=None)
        start=time.time()
        
        if pod5s != None:
            summary_file=f"{prefix}_summary.txt"
            input_name=Path(input).name
            output_fast5=f"{output}/fast5"
            if isinstance(pod5s, str):
                pod5s_dir = sorted([Path(p) for p in glob.glob(pod5s)])
            elif not pod5s:
                raise ValueError("pod5s should not be empty")
            progress.add_task(description="Pod5 to fast5...", total=None)
            if not check_path_exists(summary_file):
                convert_to_fast5_with_summary_file(pod5s_dir, output_fast5, summary_file, input_name) # type: ignore
            progress.add_task(description=f"Pod5 to fast5 Done", total=None)

            output_index=f"{input}.index"
            progress.add_task(description="Indexing reads...", total=None)
            if not check_path_exists(output_index):
                index_fastq(output_fast5, summary_file, input)
            progress.add_task(description="Indexing reads Done", total=None)

            output_ploya=f"{output}/{prefix}_ployA.tsv"
            sort_bam=f"{output}/{prefix}_ployA.sorted.bam"
            progress.add_task(description="Detecting ployA...", total=None)
            progress.add_task(description="Mapping reads to transcriptome...", total=None)
            if not check_path_exists(sort_bam):
                minimap2(input, transcriptome, sort_bam, params="-ax map-ont", threads=threads)
            progress.add_task(description="Mapping reads to transcriptome Done", total=None)
            if not check_path_exists(output_ploya):
                detect_ployA(input, sort_bam, transcriptome, output_ploya, threads=threads)
            progress.add_task(description="Detecting ployA Done", total=None)
        else:
            output_ploya=f"{output}/{prefix}_ployA.tsv"
            sort_bam=f"{output}/{prefix}_ployA.sorted.bam"
            progress.add_task(description="Detecting ployA...", total=None)
            progress.add_task(description="Mapping reads to transcriptome...", total=None)
            if not check_path_exists(sort_bam):
                minimap2(input, transcriptome, sort_bam, params="-ax map-ont", threads=threads)
            progress.add_task(description="Mapping reads to transcriptome Done", total=None)
            if not check_path_exists(output_ploya):
                ployA = ployADetector(sort_bam, output_ploya, min_a_length, max_non_a)
                ployA.analyze()
            progress.add_task(description="Detecting ployA Done", total=None)


        end=time.time()
        time_cost=f"{(end - start) // 3600}h{((end - start) % 3600) // 60}m{(end - start) % 60:.2f}s"
        print(f"Detect ployA Done, time cost: {time_cost}")
        progress.add_task(description=f"Detect ployA Done, time cost: {time_cost}", total=None)


import typer
from .modules.gene import gene
from .modules.isoform import isoform
from .modules.matrix import matrix
from .modules.polyA import polyA
from .modules.detectMod import detectMod
from .modules.isoformAS import isoformAS
from .modules.nascentRNA import nascentRNA

app = typer.Typer(add_completion=False)

@app.callback()
def callback():
    """
    nanoMD(Nanopore direct RNA sequencing Multi-dimensional analysis) was developed to synchronously analyze the changes in m6A, m5C, psi, AtoI modification sites, genes, isoforms, alternative splicing events, and nascent RNA in direct RNA sequencing data.
    """

app.command(name="gene")(gene)
app.command(name="count")(isoform)
app.command(name="polyA")(polyA)
app.command(name="matrix")(matrix)
app.command(name="isoformAS")(isoformAS)
app.command(name="detectMod")(detectMod)
app.command(name="nascentRNA")(nascentRNA)


if __name__ == "__main__":
    app()


import time, os
import typer
from typing_extensions import Annotated
from rich.progress import Progress, SpinnerColumn, TextColumn
from pathlib import Path
from basebio import check_path_exists, run_command
from ..utils.quantify import matrix_generate
from ..utils.ployA_tools import ployA_matrix_generate

app = typer.Typer()

@app.command()
def matrix(
    inputs: Annotated[str, typer.Option("--inputs", "-i", help="Regular matching pattern for salmon results directories or ployA files, such as 'path/to/*_quant' or 'path/to/*ployA.tsv'.")],
    control_names: Annotated[str, typer.Option("--control_names", "-c", help="Control sample dir name or ployA file name separated by comma, such as 'control1_quant,control2_quant,control3_quant' or 'control1_ployA.tsv,control2_ployA.tsv,control3_ployA.tsv'.")],
    prefix: Annotated[str, typer.Option("--prefix", "-p", help="Prefix for output file.")] = "prefix",
    species: Annotated[str, typer.Option("--species", "-s", help="Species name. [human|mouse|rat]")] = "human",
    type: Annotated[str, typer.Option("--type", "-t", help="Input count file type, either 'salmon' or 'ployA'.")] = "salmon",
    docker: Annotated[bool, typer.Option("--docker", help="Whether to run in docker container to plots.")] = False,
    ):
    """
    Generate matrix with salmon results or ployA files.
    """
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        try:
            progress.add_task(description="Generate count matrix and plot matrix...", total=None)
            start=time.time()

            plot_script = Path(__file__).parent.parent / "scripts"
            split_num = len(control_names.split(","))
            WKD = os.getcwd()

            if type == "salmon":
                output_count = "matrix_count.tsv"
                if not check_path_exists(output_count):
                    matrix_generate(input_dirs=inputs, control_dirs_names=control_names, output=output_count,
                                    count_type="NumReads")
                output_tpm = "matrix_tpm.tsv"
                if not check_path_exists(output_tpm):
                    matrix_generate(input_dirs=inputs, control_dirs_names=control_names, output=output_tpm,
                                    count_type="TPM")
            elif type == "ployA":
                output_count = "matrix_ployA.tsv"
                output_lengths = "ployA_lengths.tsv"
                if not check_path_exists(output_count) and not check_path_exists(output_lengths):
                    ployA_matrix_generate(input_files=inputs, control_filess_names=control_names, output_matrix=output_count, output_lengths=output_lengths)
            else:
                print(f"Error: Unknown count type {type}")
                exit(1)

            output_plot = f"matrix_plots"
            if not check_path_exists(output_plot):
                if docker:
                    run_command(f"docker run -v {WKD}:/output -v {plot_script}:/scripts -w /output legendzdy/rbase:1.0.0 Rscript /scripts/isoform.R -i /output/{output_count} -o /output/{output_plot} -p {prefix} -s {species} -t {type} -n {split_num}".split())
                else:
                    run_command(f"Rscript {plot_script}/isoform.R -i {WKD}/{output_count} -o {WKD}/{output_plot} -p {prefix} -s {species} -t {type} -n {split_num}")

            end=time.time()
            time_cost=f"{(end - start) // 3600}h{((end - start) % 3600) // 60}m{(end - start) % 60:.2f}s"
            print(f"Generate count matrix and plot matrix Done, time cost: {time_cost}")
            progress.add_task(description=f"Generate count matrix and plot matrix Done, time cost: {time_cost}", total=None)
        except Exception as e:
            print(f"Error: {e}")
            progress.add_task(description="Generate count matrix and plot matrix Failed", total=None)
            exit(1)


#!/usr/bin/env python3
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20201224
Description: This script is used to extract length and quality information of sequences form fastq file.
"""
import pandas as pd
import numpy as np
from sklearn.impute import SimpleImputer
import joblib, gzip
import argparse


def fetch_reads(sample, model):
    names = ["read", "UmisRate", "UmisQrate", "t_a_rate", "t_a_Qrate",
              "t_c_rate", "t_c_Qrate", "t_g_rate", "t_g_Qrate", "del_rate"]
    df = pd.read_csv(sample)
    df = df[names]
    df = df.replace([np.inf, -np.inf], np.nan)
    df.eval("""
    Um_UmQ = UmisRate*UmisQrate
    Um_t_a_rate = UmisRate*t_a_rate
    Um_t_a_Qrate = UmisRate*t_a_Qrate
    Um_t_c_rate = UmisRate*t_c_rate
    Um_t_c_Qrate = UmisRate*t_c_Qrate
    Um_t_g_rate = UmisRate*t_g_rate
    Um_t_g_Qrate = UmisRate*t_g_Qrate
    Um_del_rate = UmisRate*del_rate
    UQ_t_a_rate = UmisQrate*t_a_rate
    UQ_t_a_Qrate = UmisQrate*t_a_Qrate
    UQ_t_c_rate = UmisQrate*t_c_rate
    UQ_t_c_Qrate = UmisQrate*t_c_Qrate
    UQ_t_g_rate = UmisQrate*t_g_rate
    UQ_t_g_Qrate = UmisQrate*t_g_Qrate
    UQ_del_rate = UmisQrate*del_rate
    t_a_rate_t_a_Qrate = t_a_rate*t_a_Qrate
    t_a_rate_t_c_rate = t_a_rate*t_c_rate
    t_a_rate_t_c_Qrate = t_a_rate*t_c_Qrate
    t_a_rate_t_g_rate = t_a_rate*t_g_rate
    t_a_rate_t_g_Qrate = t_a_rate*t_g_Qrate
    t_a_rate_del_rate = t_a_rate*del_rate
    t_a_Qrate_t_c_rate = t_a_Qrate*t_c_rate
    t_a_Qrate_t_c_Qrate = t_a_Qrate*t_c_Qrate
    t_a_Qrate_t_g_rate = t_a_Qrate*t_g_rate
    t_a_Qrate_t_g_Qrate = t_a_Qrate*t_g_Qrate
    t_a_Qrate_del_rate = t_a_Qrate*del_rate
    t_c_rate_t_c_Qrate = t_c_rate*t_c_Qrate
    t_c_rate_t_g_rate = t_c_rate*t_g_rate
    t_c_rate_t_g_Qrate = t_c_rate*t_g_Qrate
    t_c_rate_del_rate = t_c_rate*del_rate
    t_c_Qrate_t_g_rate = t_c_Qrate*t_g_rate
    t_c_Qrate_t_g_Qrate = t_c_Qrate*t_g_Qrate
    t_c_Qrate_del_rate = t_c_Qrate*del_rate
    t_g_rate_t_g_Qrate = t_g_rate*t_g_Qrate
    t_g_rate_del_rate = t_g_rate*del_rate
    t_g_Qrate_del_rate = t_g_Qrate*del_rate
    """, inplace=True)
    test = df.iloc[:, 1:]
    imputer = SimpleImputer(missing_values=np.nan, strategy='mean')
    test_t = imputer.fit_transform(test)
    rf = joblib.load(model)
    y = rf.predict(test_t)
    dict, i = {}, 0
    while i < len(y):
        if int(y[i]) == 1:
            dict[df["read"][i]] = 1
        i += 1
    return dict

def new_fq(sample, model, rawfq, newfq):
    """
    This function is used to extract length and quality information of sequences form fastq file.

    Args:
        sample: sample raw data fecth raw signal
        model: train model
        rawfq: sample raw fastq data
        newfq: Path to an output file to be created,default= ./
    
    Example:
        new_fq(sample, model, rawfq, newfq)
    
    """
    dict_new = fetch_reads(sample, model)
    with open(newfq, 'w') as fo:
        with gzip.GzipFile(rawfq, "rb") as fi:
            while True:
                try:
                    line1 = next(fi).decode().strip()
                    line2 = next(fi).decode().strip()
                    line3 = next(fi).decode().strip()
                    line4 = next(fi).decode().strip()
                    if dict_new.get(line1.split()[0][1:], None) != None:
                        fo.write("{}\n{}\n{}\n{}\n".format(
                            line1, line2, line3, line4))
                except StopIteration:
                    break

def main():
    parser = argparse.ArgumentParser(
        description='Extract length and quality information of sequences form fastq file')
    parser.add_argument('-s', '--signal', type=str, required=False,
                        help='sample raw data fecth raw signal')
    parser.add_argument('-r', '--rawfq', type=str, required=True,
                        help='sample raw fastq data')
    parser.add_argument('-m', '--model', type=str, required=True,
                        help='train model')
    parser.add_argument('-o', '--out_file', type=str, required=True,
                        help='Path to an output file to be created,default= ./')
    args = parser.parse_args()
    new_fq(args.signal, args.model, args.rawfq, args.out_file)


if __name__ == '__main__':
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20250210
Description: This file is used to get modifications from reads.
"""

import re
import gzip
import pysam
from collections import defaultdict

class form_reads_get_modifications:
    """
    from reads get modifications
    Args:
        inputFq: input fq file
        samFile: sam file or bam file
        bedFile: single base annotaiton bed file
        output: output file
        pvalue: pvalue cutoff for modification
    """
    
    MOD_MAP = {
        'A+a': re.compile(r'A'),
        'A+17596': re.compile(r'A'),
        'T+17802': re.compile(r'T'),
        'C+m': re.compile(r'C')
    }
    
    BASE_MAP = {
        "A": ["A", "T"],
        "T": ["T", "A"],
        "C": ["C", "G"]
    }
    
    MOD_TYPE = {
        "A+a": "m6A",
        "A+17596": "AtoI",
        "T+17802": "psi",
        "C+m": "m5C"
    }

    def __init__(self, inputFq, samFile, bedFile, output, pvalue=0.98):
        self.inputFq = inputFq
        self.samFile = samFile
        self.bedFile = bedFile
        self.output = output
        self.pvalue = pvalue

    def get_annotation(self):
        def read_bed_file():
            with open(self.bedFile, 'r') as f:
                for line in f:
                    line = line.strip().split("\t")
                    chrname = f"chr{line[0].lstrip('chr')}"
                    gid = chrname + ":" + line[1]
                    enst = line[3]
                    base = line[4]
                    yield gid, (enst, base)
        return dict(read_bed_file())

    def find_poslist(self, modList):
        counter = 0
        posList = []
        for mod in modList:
            counter += int(mod)
            posList.append(counter)
            counter += 1
        return posList

    def find_read_position(self, mm, baseList, mlNum):
        modList = mm.split(",")[1:]
        posList = self.find_poslist(modList)
        mlcount = 0
        readPos = []
        readPosPvalue = []
        for pos in posList:
            rpos = baseList[pos]
            pvalue = round(int(mlNum[mlcount])/255, 3)
            readPos.append(rpos)
            readPosPvalue.append(pvalue)
            mlcount += 1
        return readPos, readPosPvalue

    def get_mod_position_with_read(self):
        rpos_mod = defaultdict(list)
        with gzip.open(self.inputFq, 'rt') as f:
                while True:
                    try:
                        name = next(f).strip()
                        seq = next(f).strip()
                        plus = next(f).strip()
                        qual = next(f).strip()
                        
                        namelist = name.split("\t")
                        id = namelist[0].split("@")[-1]
                        MMtagList = namelist[-2].split("MM:Z:")[-1].split(";")
                        MLtag = namelist[-1].split(",")[1:]
                        mlstart = 0
                        for mm in MMtagList:
                            mmlen = len(mm.split(","))
                            if mmlen > 1:
                                mlend = mlstart + mmlen - 1
                                mlNum = MLtag[mlstart:mlend]
                                mlstart += mmlen - 1
                            for mod, base in self.MOD_MAP.items():
                                if mm.startswith(mod):
                                    base_list = [match.start() + 1 for match in base.finditer(seq)]
                                    readPos, readPosPvalue = self.find_read_position(mm, base_list, mlNum)
                                    for rpos, pvalue in zip(readPos, readPosPvalue):
                                        if pvalue >= self.pvalue:
                                            rpos_mod[id].append((rpos, pvalue, mod))
                    except StopIteration:
                        break
        return rpos_mod

    def get_mod_position_with_sam(self):
        rpos_mod = self.get_mod_position_with_read()
        annot = self.get_annotation()
        with open(self.output, "w") as f:
            for read in pysam.AlignmentFile(self.samFile, "r"):
                if not read.is_unmapped and rpos_mod.get(read.query_name) is not None:
                    chrname = f"chr{read.reference_name.lstrip('chr')}"
                    pos = read.reference_start
                    read_pos = 1
                    strand = "-" if read.is_reverse else "+"
                    for op, length in read.cigartuples:
                        if op == 0:
                            for i in range(length):
                                mod_list = rpos_mod[read.query_name]
                                for rpos, pvalue, mod in mod_list:
                                    if rpos == read_pos + i:
                                        genome_id = f"{chrname}:{pos + i}"
                                        if annot.get(genome_id) is not None:
                                            enst, base = annot[genome_id]
                                            rbase = mod[0]
                                            mod_type = self.MOD_TYPE[mod]
                                            if rbase in self.BASE_MAP and base in self.BASE_MAP[rbase]:
                                                f.write(f"{chrname}\t{pos + i}\t{pos + i + 1}\t{pvalue}\t{enst}\t{strand}\t{mod_type}\t{base}\n")
                                            else:
                                                pass
                                                # print(f"rbase: {rbase}, base: {base} not supported, {self.BASE_MAP}")
                                        else:
                                            pass
                                            # print(f"genome_id: {genome_id} not found in annotation file")
                            pos += length
                            read_pos += length
                        elif op == 1:  # I: 插入
                            # 插入的碱基在参考基因组上不占位置，但在 read 上有位置
                            read_pos += length
                        elif op == 2:  # D: 删除
                            # 删除的碱基在参考基因组上仍占位置，但跳过它们
                            pos += length
                        elif op == 4:  # S: 跳过
                            # 跳过的碱基不占基因组位置，但在 read 上有位置
                            read_pos += length
                        elif op == 5:  # H: 硬剪切
                            # 硬剪切不占基因组位置
                            read_pos += length
                        else:
                            print(f"op: {op}, length: {length} not supported")
            else:
                print(f"read: {read.query_name} not found in fq file")


# -*- coding: utf-8 -*-
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20250612
Description: This file contains the functions for converting pod5 files to fast5 files and indexing the fastq file using nanopolish index.
"""
from concurrent.futures import Future, ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import Dict, List
from more_itertools import chunked

import pysam, os, glob
import pod5 as p5
import pandas as pd
from pod5.tools.utils import DEFAULT_THREADS, collect_inputs, limit_threads
from pod5.tools.pod5_convert_to_fast5 import * # type: ignore
from basebio import run_command

def convert_to_fast5_with_summary_file(
    inputs: List[Path],
    output: Path,
    summary_file: str,
    fastq_name: str,
    recursive: bool = False,
    threads: int = DEFAULT_THREADS,
    force_overwrite: bool = False,
    file_read_count: int = 4000,
):
    if output.exists() and not output.is_dir():
        raise FileExistsError("Cannot output to a file")

    threads = limit_threads(threads)
    columns = ["filename_fastq", "filename_fast5", "read_id", "run_id", "channel", "mux", "start_time", "duration", "num_events", "passes_filtering", "template_start", "num_events_template", "template_duration", "sequence_length_template", "mean_qscore_template", "strand_score_template", "median_template", "mad_template", "pore_type", "experiment_id", "sample_id", "end_reason", "alias", "type", "barcode_arrangement", "barcode_full_arrangement", "barcode_kit", "barcode_variant", "barcode_score", "barcode_front_id", "barcode_front_score", "barcode_front_refseq", "barcode_front_foundseq", "barcode_front_foundseq_length", "barcode_front_begin_index", "barcode_rear_id", "barcode_rear_score", "barcode_rear_refseq", "barcode_rear_foundseq", "barcode_rear_foundseq_length", "barcode_rear_end_index"]

    # 数据行（单行列表）
    data_row = ["test", "885", "1", "56.132750", "0.756500", "605", "TRUE", "56.154000", "588", "0.735250", "153", "8.314826", "0.000000", "94.303551", "8.406906", "not_set", "20220426-CDS0204-PAK00515", "no_sample", "signal_positive", "barcode11", "test_sample", "barcode11", "LWB11_var1", "LWB", "var1", "92.750000", "BC11_FWD", "89.250000", "CCGTGACGTT", "51", "34", "BC11_REV", "92.750000", "GCAATAA", "GCACATC", "52", "20"]
    
    with open(summary_file, "w") as fo:
        fo.write("\t".join(columns) + "\n")
        with ProcessPoolExecutor(max_workers=threads) as executor:
            total_reads = 0
            futures: Dict[Future, Path] = {}
            # Enumerate over input pod5 files
            for input_idx, source in enumerate(
                collect_inputs(inputs, recursive, "*.pod5", threads=threads)
            ):
                # Open the inputs to read the read ids
                with p5.Reader(source) as reader:
                    for chunk_idx, read_ids in enumerate(
                        chunked(reader.read_ids, file_read_count)
                    ):
                        dest = (
                            output / f"{source.stem}.{chunk_idx}_{input_idx}.fast5"
                        ).resolve()

                        if dest.exists() and not force_overwrite:
                            raise FileExistsError(
                                "Output path points to an existing file and --force-overwrite not set"
                            )

                        kwargs = {
                            "source": source,
                            "dest": dest,
                            "read_ids": read_ids,
                        }
                        futures[executor.submit(convert_pod5_to_fast5, **kwargs)] = dest  # type: ignore

                        for read_id in read_ids:
                            raw = [fastq_name, f"{source.stem}.{chunk_idx}_{input_idx}.fast5", read_id] + data_row
                            fo.write("\t".join(raw) + "\n")

                    total_reads += len(reader.read_ids)

        print(f"Converting pod5s into {len(futures)} fast5 files. Please wait...")

        status = StatusMonitor(file_count=len(inputs))
        status.increment(files_started=len(inputs), read_count=total_reads)

        for idx, future in enumerate(as_completed(futures)):
            (reads_converted, samples_converted) = future.result()

            status.increment(
                files_ended=1,
                sample_count=samples_converted,
                reads_processed=reads_converted,
            )
            status.print_status()

        status.print_status(force=True)

        print("Conversion complete")

def index_fastq(fast5_dir, summary_file, fastq):
    """
    Index the fastq file using nanopolish index.
    
    Args:
        fast5_dir (str): The directory where the fast5 files are located.
        summary_file (str): The path to the sequencing summary file.
        fastq (str): The path to the fastq file.
    
    Examples:
        index_fastq("fast5_dir", "summary.txt", "fastq.gz")
    """

    cmd = f"nanopolish index --directory={fast5_dir} --sequencing-summary={summary_file} {fastq}".split()
    run_command(cmd)

def detect_ployA(fastq, bam, transcriptome, output_ploya, threads=8):
    """
    Detect ployA using nanopolish ployA.

    Args:
        fastq (str): The path to the fastq file.
        bam (str): The path to the bam file.
        transcriptome (str): The path to the transcriptome file.
        output_ploya (str): The path to the output ployA file.
        threads (int): The number of threads to use.

    Examples:
        detect_ployA("fastq.gz", "bam", "transcriptome.fa", "ployA.tsv", threads=8)
    """

    cmd = f"nanopolish ployA --threads={threads} --reads={fastq} --bam={bam} --genome={transcriptome} > {output_ploya}"
    run_command(cmd, use_shell=True)

class ployADetector:
    """
    ployA detector.

    Args:
        bam_path (str): The path to the bam file.
        output_path (str): The path to the output file.
        min_a_length (int): The minimum length of ployA.
        max_non_a (int): The maximum number of non-A bases allowed in ployA.
    
    Examples:
        detector = ployADetector("bam", "ployA.tsv", min_a_length=12, max_non_a=3)
        detector.analyze()
    """
    def __init__(self, bam_path, output_path, min_a_length=12, max_non_a=3):
        self.bam_path = bam_path
        self.output_path = output_path
        self.min_a_length = min_a_length
        self.max_non_a = max_non_a
    
    @staticmethod
    def reverse_complement(seq):
        """
        Get the reverse complement of a sequence.
        """
        complement = {'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G', 'U': 'A',
                    'a': 't', 't': 'a', 'g': 'c', 'c': 'g', 'u': 'a',
                    'N': 'N', 'n': 'n'}
        return ''.join(complement.get(base, base) for base in reversed(seq)) # type: ignore
    
    def find_longest_ployA(self, seq):
        """
        Use sliding window algorithm to find the longest consecutive A sequence, allowing a few non-A bases.
        """
        if not seq:
            return "", 0, 0, 0.0
        
        n = len(seq)
        left = 0
        non_a_count = 0
        max_length = 0
        a_count_in_max = 0
        best_start = 0
        best_end = -1
        
        for right in range(n):
            if seq[right].upper() != 'A':
                non_a_count += 1
            
            while non_a_count > self.max_non_a:
                if seq[left].upper() != 'A':
                    non_a_count -= 1
                left += 1
            
            current_length = right - left + 1
            current_a_count = current_length - non_a_count

            if current_length > max_length:
                max_length = current_length
                a_count_in_max = current_a_count
                best_start = left
                best_end = right
        
        a_ratio = a_count_in_max / max_length if max_length > 0 else 0.0
        
        ployA_seq = seq[best_start:best_end+1] if max_length > 0 else ""
        
        return ployA_seq, max_length, a_count_in_max, a_ratio
    
    def process_read(self, read):
        if read.is_unmapped or not read.cigartuples:
            return None
        
        read_name = read.query_name
        seq = read.query_sequence
        if not seq:
            return None
            
        flag = read.flag
        cigar = read.cigartuples
        
        if flag == 0:
            target_op = cigar[-1]
        elif flag == 16:
            target_op = cigar[0]
        else:
            return None
        
        op_type, op_len = target_op
        
        if op_type != 4:
            return None
        
        if flag == 0:
            target_seq = seq[-op_len:] if op_len <= len(seq) else seq
        elif flag == 16:
            target_seq = self.reverse_complement(seq[:op_len]) if op_len <= len(seq) else seq
        else:
            return None
        
        ployA_seq, total_length, a_count, a_ratio = self.find_longest_ployA(target_seq)
        
        has_ployA = a_count >= self.min_a_length and a_ratio >= 0.75
        
        return {
            "read_name": read_name,
            "ref_name": read.reference_name,
            "strand": "+" if flag == 0 else "-",
            "ployA_seq": ployA_seq,
            "ployA_region_length": total_length,
            "a_count": a_count,
            "a_ratio": a_ratio,
            "has_ployA": "Yes" if has_ployA else "No"
        }
    
    def analyze(self):
        with open(self.output_path, 'w') as fout:
            fout.write("readName\trefName\tstrand\tployASeq\tployALength\tACount\tARatio\tHasployA\n")
            with pysam.AlignmentFile(self.bam_path, "rb") as bam:
                for read in bam:
                    result = self.process_read(read)
                    if result:
                        if result["has_ployA"] == "Yes":
                            fout.write("\t".join([
                                result["read_name"],
                                result["ref_name"],
                                result["strand"],
                                result["ployA_seq"],
                                str(result["ployA_region_length"]),
                                str(result["a_count"]),
                                f"{result['a_ratio']:.3f}",
                                result["has_ployA"]
                            ]) + "\n")

def drop_outliers(df, dep_col, handle_nan='keep'):
    """
    Use IQR method to drop outliers from DataFrame

    Args:
        df: pandas DataFrame
        dep_col: str, the column name for detecting outliers
        handle_nan: str, the method for handling NaN values ('keep','remove', 'fill')
    
    Returns:
        The filtered DataFrame
    """

    # Validate column exists
    if dep_col not in df.columns:
        raise ValueError(f"Column '{dep_col}' not found in DataFrame")

    # Handle NaN values
    if handle_nan == 'remove':
        df = df.dropna(subset=[dep_col])
    elif handle_nan == 'fill':
        median_val = df[dep_col].median()
        df[dep_col] = df[dep_col].fillna(median_val)

    # Default 'keep' does not do anything
    # Calculate quartiles and IQR
    a = df[dep_col]
    q1 = a.quantile(0.25)
    q3 = a.quantile(0.75)
    iqr = q3 - q1

    # Calculate normal value range
    bottom = q1 - 1.5 * iqr
    top = q3 + 1.5 * iqr

    # Create boolean mask
    mask = (a >= bottom) & (a <= top)
    
    return df[mask]

def ployA_matrix_generate(input_files, control_filess_names, output_matrix, output_lengths):
    """
    Generate ployA matrix and ployA lengths file.
    """
    if isinstance(input_files, str):
        sorted_files = sorted(glob.glob(input_files))
        if not sorted_files:
            raise ValueError(f"No matching file found for {input_files}")
        input_path = os.path.dirname(sorted_files[0])
        control_files = [os.path.join(input_path, control_name) for control_name in control_filess_names.split(',')]
        for control_file in control_files:
            if control_file in sorted_files:
                sorted_files.remove(control_file)   
            else:
                raise ValueError(f"Control file {control_file} not found in {input_files}")
        input_files = control_files + sorted_files
    elif not input_files:
        raise ValueError("input_files should not be empty")
    
    # Initialize a dictionary to store all counts
    all_counts = {}
    sample_names = []
    # 创建一个空的DataFrame存储所有结果
    all_ployA_lengths = pd.DataFrame()
    
    # Process each sample file
    for sample_file in input_files:
        sample_name = os.path.basename(sample_file).split('.')[0].replace('_ployA', '')
        sample_names.append(sample_name)
        condition = sample_name
        
        # Read the TSV file
        df = pd.read_csv(sample_file, sep='\t')
        
        # Count occurrences of each refName
        ref_counts = df['refName'].value_counts().to_dict()
        all_counts[sample_name] = ref_counts
        # ployA lengths
        df_temp = df[['ployALength']].copy()
        df_temp['condition'] = condition
        df_temp = df_temp.rename(columns={'ployALength': 'measurement'})
        all_ployA_lengths = pd.concat([all_ployA_lengths, df_temp], ignore_index=True)
    
    # Create a set of all unique refNames
    all_refnames = set()
    for counts in all_counts.values():
        all_refnames.update(counts.keys())
    
    # Create the matrix DataFrame
    matrix = pd.DataFrame(index=list(all_refnames), columns=sample_names)
    matrix = matrix.fillna(0)  # Fill with zeros initially
    # Populate the matrix with counts
    for sample, counts in all_counts.items():
        for refname, count in counts.items():
            matrix.at[refname, sample] = count
    # Save the matrix to output file
    matrix.to_csv(output_matrix, sep='\t')

    # ployA lengths
    all_ployA_lengths = all_ployA_lengths[['condition', 'measurement']]
    all_ployA_lengths = drop_outliers(all_ployA_lengths, 'measurement')
    all_ployA_lengths.to_csv(output_lengths, sep='\t', index=False)


# -*- coding: utf-8 -*-
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20250612
Description: This file contains the quantification function.
"""
import os, glob
import pandas as pd
from collections import OrderedDict
from basebio import run_command, minimap2

def salmon_map(input, reference, output):
    """
    Map with minimap2.
    Args:
        input: input fasta file.
        reference: reference transcripts fasta file.
        output: output bam file.
    """
    tool = "minimap2"
    parms = "--secondary=no --cs -a"
    threads = 4
    minimap2(input, reference, output, tool, parms, threads)
    

def salmon_quantify(input, reference, output):
    """
    Quantify with salmon.
    Args:
        input: input bam file.
        reference: reference transcripts fasta file.
        output: output quantification file.
    """
    
    command = [
        "salmon", "quant", "--noErrorModel", "-p", "4",
        "-t", reference, "-l", "SF", 
        "-a", input, "-o", output
        ]
    run_command(command)

def matrix_generate(input_dirs, control_dirs_names, output, count_type):
    """
    Generate count matrix from salmon results.
    Args:
        input_dirs:  Regular matching pattern for salmon results directories, such as 'path/to/*_quant'.
        output: output file name.
        count_type: count type to extract ("NumReads" or "TPM").
    """

    if count_type not in ["NumReads", "TPM"]:
        raise ValueError("count_type should be 'NumReads' or 'TPM'")
    
    count_data = OrderedDict()
    transcript_ids = None

    if isinstance(input_dirs, str):
        sorted_dirs = sorted(glob.glob(input_dirs))
        if not sorted_dirs:
            raise ValueError(f"No matching directory found for {input_dirs}")
        input_path = os.path.dirname(sorted_dirs[0])
        control_dirs = [os.path.join(input_path, control_name) for control_name in control_dirs_names.split(',')]
        for control_dir in control_dirs:
            if control_dir in sorted_dirs:
                sorted_dirs.remove(control_dir)
            else:
                raise ValueError(f"Control directory {control_dir} not found in {input_dirs}")
        input_dirs = control_dirs + sorted_dirs
    elif not input_dirs:
        raise ValueError("input_dirs should not be empty")
    
    for sample_dir in input_dirs:
        sample_name = os.path.basename(os.path.normpath(sample_dir))
        quant_file = os.path.join(sample_dir, "quant.sf")
        
        if not os.path.exists(quant_file):
            raise FileNotFoundError(f"quant.sf not found in {sample_dir}")
        
        df = pd.read_csv(quant_file, sep='\t')
        
        if "Name" not in df.columns or count_type not in df.columns:
            raise ValueError(f"Required columns missing in {quant_file}")
        
        if transcript_ids is None:
            transcript_ids = df["Name"].tolist()
            count_data["transcript_id"] = transcript_ids
        else:
            if transcript_ids != df["Name"].tolist():
                raise ValueError("Transcript IDs mismatch between samples. "
                                 "Ensure all samples used the same reference.")
        count_data[sample_name] = df[count_type].tolist()
    
    count_df = pd.DataFrame(count_data)
    count_df.set_index("transcript_id", inplace=True)
    count_df.to_csv(output, sep='\t')


# -*- coding: utf-8 -*-
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20250218
Description: This file is used to calculate the relative and absolute distances from UTR and CDS endpoints.
"""
class gene_feature_distance_calculator:
    """
    calculate the relative and absolute distances from UTR and CDS endpoints
    args:
        bed_file: path to the BED file
        regions_file: path to the UTR/CDS regions file
        output: path to the output file
    output:
        a bed file with the following columns:
            chr: chromosome
            coord: coordinate
            gene_name: gene name
            refseqID: refseq ID
            rel_location: relative distance from 5'UTR, CDS, or 3'UTR
            utr5_st: start position of 5'UTR
            utr5_end: end position of 5'UTR
            cds_st: start position of CDS
            cds_end: end position of CDS
            utr3_st: start position of 3'UTR
            utr3_end: end position of 3'UTR
            utr5_size: size of 5'UTR
            cds_size: size of CDS
            utr3_size: size of 3'UTR
    """
    def __init__(self, bed_file, regions_file, output):
        """初始化，加载 BED 文件和 UTR/CDS 文件"""
        self.bed_file = bed_file
        self.regions_file = regions_file
        self.refseq_endpts = self.load_refseq_endpoints()
        self.output = output

    def load_refseq_endpoints(self):
        """加载 5'UTR, CDS, 3'UTR 的坐标数据"""
        refseq_endpts = {}
        with open(self.regions_file, 'r') as f:
            for line in f:
                if "\tNA" in line:  # Excludes non-coding RNA (they lack 5'UTR, CDS, and 3'UTR annotations)
                    continue
                parts = line.strip().split("\t")
                refseq = parts[0]
                refseq_endpts[refseq] = list(map(int, parts[3:]))  # Store the start/end points as integers
        return refseq_endpts

    def abs_distance(self, mrna_pos, endpts):
        """计算 mRNA 相对于每个区域的绝对距离"""
        return [mrna_pos - point for point in endpts]

    def rel_distance(self, mrna_pos, endpts):
        """计算 mRNA 相对于 5'UTR, CDS 和 3'UTR 的相对位置"""
        u5_st, u5_end, cds_st, cds_end, u3_st, u3_end = endpts
        if u5_st <= mrna_pos <= u5_end:
            return (mrna_pos - u5_st + 1) / (u5_end - u5_st + 1)
        elif cds_st <= mrna_pos <= cds_end:
            return (mrna_pos - cds_st + 1) / (cds_end - cds_st + 1) + 1
        elif u3_st <= mrna_pos <= u3_end:
            return (mrna_pos - u3_st + 1) / (u3_end - u3_st + 1) + 2
        return None  # if mrna_pos is outside of all regions

    def size_features(self, endpts):
        """计算 5'UTR, CDS 和 3'UTR 的大小"""
        u5_st, u5_end, cds_st, cds_end, u3_st, u3_end = endpts
        return [u5_end - u5_st, cds_end - cds_st, u3_end - u3_st]

    def process_bed_file(self):
        """处理 BED 文件，计算每个基因的相关信息"""
        excluded = 0
        total = 0

        with open(self.bed_file, 'r') as f , open(self.output, 'w') as out:
            out.write("chr\tcoord\tgene_name\trefseqID\trel_location\tutr5_st\tutr5_end\tcds_st\tcds_end\tutr3_st\tutr3_end\tutr5_size\tcds_size\tutr3_size\n")
            for line in f:
                if line.startswith("#"):  # Skip comment lines
                    continue
                total += 1
                parts = line.strip().split("\t")
                mrna_meta = parts[4]
                refseq, gname, region, mrna_pos = mrna_meta.split("|")

                if refseq not in self.refseq_endpts:
                    excluded += 1
                    continue

                endpts = self.refseq_endpts[refseq]
                mrna_pos = int(mrna_pos)

                abs_dist = self.abs_distance(mrna_pos, endpts)
                rel_dist = self.rel_distance(mrna_pos, endpts)
                feature_sizes = self.size_features(endpts)

                out.write(f"{parts[0]}\t{parts[2]}\t{gname}\t{refseq}\t{rel_dist}\t" + "\t".join(map(str, abs_dist)) + "\t" + "\t".join(map(str, feature_sizes)) + "\n")

        print(f"** Total bedfile lines: {total}")
        print(f"** Excluded (features not well defined or gene not found in gtf): {excluded}")
